import Flutter
import UIKit
import LinkPresentation
import Microfutures

var shareId = 0
public class SwiftFlutterShareContentPlugin: NSObject, FlutterStreamHandler, FlutterPlugin {
    public func onListen(withArguments arguments: Any?,
                         eventSink events: @escaping FlutterEventSink) -> FlutterError? {
        self.events = events
        return nil
    }
    
    public func onCancel(withArguments arguments: Any?) -> FlutterError? {
        self.events = nil
        return nil
    }
    
    var events: FlutterEventSink? = nil
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        
        let channel = FlutterMethodChannel(name: "github://SunnyApp/flutter_share_content",
                                           binaryMessenger: registrar.messenger())
        let instance = SwiftFlutterShareContentPlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
        let events = FlutterEventChannel(name: "github://SunnyApp/flutter_share_content_events",
                                         binaryMessenger: registrar.messenger())
        events.setStreamHandler(instance)
        
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch(call.method) {
        case "share":
            
            guard let args = call.arguments as? [String:Any] else {
                result(FlutterError.init(code: "invalid", message: "Expected an object", details: nil))
                return
            }
            guard let items = args["items"] as? [[String:Any]] else {
                result(FlutterError.init(code: "invalid.items", message: "Invalid items", details: nil))
                return
            }
            guard let allowedActions = args["allowedActions"] as? [[String:String]]? else {
                result(FlutterError.init(code: "invalid.allowedActions", message: "Invalid share media", details: nil))
                return
            }
            guard let restrictedActions = args["restrictedActions"] as? [[String:String]]? else {
                result(FlutterError.init(code: "invalid.allowedActions", message: "Invalid share media", details: nil))
                return
            }
            shareId += 1
            let id = "\(shareId)"
            
            DispatchQueue.global().async {
                do {
                    let items = try items.compactMap{ try ShareItem.init($0)}
                    DispatchQueue.main.async {
                        do {
                            self.openShare(shareId: id,
                                           restrictedActions: try restrictedActions?.compactMap{ try TargetAction.init($0) },
                                           allowedActions: try allowedActions?.compactMap{ try TargetAction.init($0) },
                                           items: items)
                            result(["shareId": id])
                        } catch FlutterException.invalid(let code, let message) {
                            result(FlutterError.init(code: code, message: message, details: nil))
                        } catch let error  {
                            result(FlutterError.init(code: "unknownError", message: "\(error)", details: nil))
                        }
                    }
                } catch FlutterException.invalid(let code, let message) {
                    result(FlutterError.init(code: code, message: message, details: nil))
                } catch let error  {
                    result(FlutterError.init(code: "unknownError", message: "\(error)", details: nil))
                }
            }
            
        default:
            result(FlutterError.init(code: "unknownOperation", message: "This is not supported", details: nil))
        }
    }
    
    func openShare(shareId: String, restrictedActions: [TargetAction]?, allowedActions: [TargetAction]?, items: [ShareItem]) {
        
        //let content = ShareContent.init(shareId: shareId, items: items)
        let controller: UIViewController = UIApplication.shared.keyWindow!.rootViewController!
        
        let ac = UIActivityViewController.init(activityItems: items,
                                               applicationActivities:nil)
        
        ac.popoverPresentationController?.sourceView = controller.view
        ac.completionWithItemsHandler = {type,completed,result,err in
            if let event = self.events {
                event([
                    "shareId": shareId,
                    "activityType": type?.rawValue,
                    "isCompleted": completed,
                    "error": String(describing: err),
                    "result": String(describing: result),
                    ] as [String: Any?])
            }
        }
        /// wait for all items to full load.
        
        controller.present(ac, animated:true)
    }
}

struct ShareMetadata{
    let icon:String?
    let imageUrl:String?
    let caption:String?
    let videoUrl:String?
    
    init?(_ from: Any?) throws {
        guard let nn = from else {
            return nil
        }
        guard let from = nn as? [String:Any?]  else {
            throw FlutterException.invalid(code: "invalidMetadata", message: "Expected a different format")
        }
        self.icon = from["icon"] as? String
        self.imageUrl = from["imageUrl"] as? String
        self.caption = from["caption"] as? String
        self.videoUrl = from["videoUrl"] as? String
        
    }
}

struct ShareVariant {
    let content: Any
    let rawContent: Any?
    let targetAction: TargetAction
    let shareType: ShareType
    let subject: String?
    
    init?(_ from: [String:Any?]) throws {
        
        if let targetAction = try TargetAction.init(from["targetAction"] ?? nil) {
            self.targetAction = targetAction
        } else {
            throw  FlutterException.invalid(code: "arg.targetAction", message: "targetAction is missin")
        }
        
        let shareType = ShareType.init(any: from["shareType"] ?? nil)
        let rawContent = from["content"] ?? nil
        self.shareType = shareType
        self.rawContent = rawContent
        self.subject = from["subject"] as? String
        self.content = try shareType.convertValue(value: rawContent)
    }
    
    // Determines whether this shareItem applies to the given activityType
    func appliesTo(_ activityType: UIActivity.ActivityType?)->Bool {
        return targetAction.appliesTo(activityType)
    }
    
    var variantContent:Any {
        get{
            return content
        }
    }
}

func loadShareItem(_ from: Any?)->Future<ShareItem?> {
    return Future { completion in
        DispatchQueue.global().async {
            do {
                let item = try ShareItem.init(from)
                completion(.success(item))
            } catch let error {
                completion(.failure(error))
            }
        }
    }
}

class ShareItem: NSObject, UIActivityItemSource {
    
    let rawContent: Any?
    let content: Any
    let label: String?
    let subject: String?
    let shareType:ShareType
    let metadata: ShareMetadata?
    let variants: [ShareVariant]?
    
    init?(_ from: Any?) throws {
        guard let from = from as? [String:Any] else {
            return nil
        }
        
        let shareType = ShareType.init(rawValue: from["shareType"] as! Int)
        let rawContent = from["content"] ?? nil
        self.shareType = shareType
        self.rawContent = rawContent
        self.content =  try shareType.convertValue(value: rawContent)
        self.metadata = try ShareMetadata.init(from["metadata"])
        self.subject = from["subject"] as? String
        self.label = from["label"] as? String
        if let variants = from["variants"] as? [String: [String:Any?]] {
            self.variants = try variants.values.compactMap { try ShareVariant.init($0) }
        } else {
            self.variants = [ShareVariant]()
        }
    }
    
    // Finds the appropriate ShareItem for the given activityType, or self if no variant is found
    private func getContentForActivityType(_ activityType: UIActivity.ActivityType?)->Any {
        let variant = getVariant(activityType)
        return variant?.variantContent ?? self.content
    }
    
    // Finds the appropriate ShareItem for the given activityType, or self if no variant is found
    private func getVariant(_ activityType: UIActivity.ActivityType?)->ShareVariant? {
        guard let activityType = activityType else {
            // We were passed a null activityType - use ourself as default
            return nil
        }
        
        guard let variants = self.variants else {
            // We don't have variants
            return nil
        }
        
        let variant = variants.filter { $0.appliesTo(activityType) }.first
        return variant
    }
    
    
    func activityViewControllerPlaceholderItem(_ activityViewController: UIActivityViewController) -> Any {
        return self.content
    }
    
    func activityViewController(_ activityViewController: UIActivityViewController, itemForActivityType activityType: UIActivity.ActivityType?) -> Any? {
        return getContentForActivityType(activityType)
    }
    
    func activityViewController(_ activityViewController: UIActivityViewController, subjectForActivityType activityType: UIActivity.ActivityType?) -> String {
        let variant = getVariant(activityType)
        return variant?.subject ?? self.subject ?? self.label ?? ""
    }
    
    @available(iOS 13.0, *)
    func activityViewControllerLinkMetadata(_: UIActivityViewController) -> LPLinkMetadata? {
        guard let md = self.metadata else {
            return nil
        }
        
        let metadata = LPLinkMetadata()
        if let label = self.label {
            metadata.url = URL(fileURLWithPath: label)
        }
        //metadata.originalURL = selfContent
        if let caption = md.caption {
            metadata.originalURL = URL(fileURLWithPath: caption)
        }
        
        metadata.title = self.label ?? String(describing: self.content)
        
        if let videoUrl = md.videoUrl {
            metadata.remoteVideoURL = URL(string: videoUrl)
        }
        if let imageUrl = md.imageUrl {
            DispatchQueue.global().async {
                if let data = try? Data(contentsOf: URL(string: imageUrl)! ) {
                    
                    if let image = UIImage(data: data) {
                        DispatchQueue.main.async {
                            metadata.imageProvider = NSItemProvider.init(object: image)
                        }
                    }
                }
                
            }
        }
        
        if let iconUrl = md.icon {
            DispatchQueue.global().async {
                if let data = try? Data(contentsOf: URL(string: iconUrl)! ) {
                    
                    if let image = UIImage(data: data) {
                        DispatchQueue.main.async {
                            metadata.iconProvider = NSItemProvider.init(object: image)
                        }
                    }
                }
                
            }
            
            
        }
        
        return metadata
    }
}

enum FlutterException: Error {
    case invalid(code: String, message: String)
}

// Targets a specific application action with specialized content
struct TargetAction {
    init?(_ values: Any?) throws {
        guard let values = values as? [String:Any?] else {
            return nil
        }
        guard let key = values["key"] as? String else {
            throw FlutterException.invalid(code: "invalid.targetAction.key", message: "Invalid key")
        }
        switch (key) {
        case "postToFacebook": self.key = UIActivity.ActivityType.postToFacebook
        case "shareWithDevice": self.key = UIActivity.ActivityType.airDrop
        case "sendEmail": self.key = UIActivity.ActivityType.mail
        case "sendSms": self.key = UIActivity.ActivityType.message
        case "postToTencentWeibo": self.key = UIActivity.ActivityType.postToTencentWeibo
        case "postToVimeo": self.key = UIActivity.ActivityType.postToVimeo
        case "postToFlickr": self.key = UIActivity.ActivityType.postToFlickr
        case "postToTwitter": self.key = UIActivity.ActivityType.postToTwitter
        case "postToWeibo": self.key = UIActivity.ActivityType.postToWeibo
        default: self.key = nil
        }
        
        self.iosName = values["iosName"] as? String
    }
    let key: UIActivity.ActivityType?
    let iosName: String?
    
    func appliesTo(_ activityType: UIActivity.ActivityType?)->Bool {
        
        guard let activityType = activityType else {
            return true
        }
        
        if let iosName = iosName {
            return activityType.rawValue == iosName
        }
        
        guard let key = self.key else {
            return false
        }
        return key.rawValue == activityType.rawValue
    }
}

enum ShareType: Int {
    init(rawValue: Int) {
        switch(rawValue) {
        case 0: self = .image
        case 1: self = .text
        case 2: self = .html
        case 3: self = .url
        case 4: self = .file
        default: self = .unknown
        }
    }
    
    init(any: Any?) {
        if let any = any as? Int {
            switch(any) {
            case 0: self = .image
            case 1: self = .text
            case 2: self = .html
            case 3: self = .url
            case 4: self = .file
            default: self = .unknown
            }
        } else{
            self = .unknown
        }
        
    }
    
    case image = 0
    case text = 1
    case html = 2
    case url = 3
    case file = 4
    case unknown
}

func loadContentFuture(shareType: ShareType, rawContent: Any?) -> Future<Any> {
    return Future { completion in
        DispatchQueue.global().async {
            do {
                let content = try shareType.convertValue(value: rawContent)
                completion(.success(content))
            } catch let error {
                completion(.failure(error))
            }
        }
    }
}

extension ShareType {
    /// This may load from the network, so it should happen asynchronously
    func convertValue(value: Any?)throws ->Any {
        guard let value = value else {
            throw FlutterException.invalid(code: "arg.content", message: "Content can't be null")
        }
        switch(self) {
            
        case .image:
            return try asImage(value)
        case .text:
            return try asString(value)
        case .html:
            return try asString(value)
        case .url:
            return try asURL(value)
        case .file:
            return try asFile(value)
        case .unknown:
            return value
        }
    }
}


func asFile(_ self:Any) throws -> Any {
    throw FlutterException.invalid(code: "notSupported", message: "File types not supported")
}

func asImage(_ self:Any) throws -> UIImage {
    guard let imageUrl = self as? String else {
        throw FlutterException.invalid(code: "invalidImage", message: "Expected an image URL")
    }
    
    var image: UIImage? = nil
    
    
    if let data = try? Data(contentsOf: URL(string: imageUrl)! ) {
        image = UIImage(data: data)
    }
    
    if let result = image {
        return result
    } else {
        throw FlutterException.invalid(code: "invalidImage", message: "Tried to read image from \(imageUrl) but failed")
    }
}

func asURL(_ self:Any) throws -> URL {
    guard let url = self as? URL else {
        guard let newUrl = URL(string: try asString(self)) else {
            throw FlutterException.invalid(code: "invalidURL", message: "Unable to construct URL from \(try asString(self))")
        }
        return newUrl
    }
    return url
}

func asString(_ self:Any) throws -> String {
    guard let str = self as? String else {
        throw FlutterException.invalid(code: "invalidType", message: "Could not convert value to String")
    }
    return str
}

