import 'package:flutter/foundation.dart';

enum ShareType { image, text, html, url, file }

abstract class ShareContent {
  dynamic get content;

  ShareType get shareType;

  // optional
  String? get subject;

  factory ShareContent.image(String imageUrl, {String? subject}) =>
      _ShareContent(imageUrl, ShareType.image, subject: subject);

  factory ShareContent.file(file, {String? subject}) =>
      _ShareContent(file, ShareType.file, subject: subject);

  factory ShareContent.text(String text, {String? subject}) =>
      _ShareContent(text, ShareType.text, subject: subject);

  factory ShareContent.html(String html, {String? subject}) =>
      _ShareContent(html, ShareType.html, subject: subject);

  factory ShareContent.url(String url, {String? subject}) =>
      _ShareContent(url, ShareType.url, subject: subject);
}

class _ShareContent implements ShareContent {
  @override
  final dynamic content;

  @override
  final ShareType shareType;

  @override
  final String? subject;
  const _ShareContent(this.content, this.shareType, {this.subject});
}

class ShareItem {
  final dynamic content;
  final ShareType shareType;
  final String? subject;
  final String? label;
  final ShareMetadata? metadata;
  final Map<String, TargetedShareContent> variants;

  ShareItem._({
    this.content,
    this.subject,
    this.metadata,
    required this.shareType,
    this.label,
    Map<String, TargetedShareContent>? variants,
  }) : variants = <String, TargetedShareContent>{
          ...?variants,
        };

  factory ShareItem.image(String imageUrl,
      {String? subject, String? caption, String? label}) {
    return ShareItem._(
        content: imageUrl,
        subject: subject,
        label: label,
        shareType: ShareType.image,
        metadata: ShareMetadata(
          caption: caption,
          imageUrl: imageUrl,
        ));
  }

  factory ShareItem.url(String url,
      {String? caption,
      String? subject,
      String? label,
      String? imageUrl,
      String? iconUrl,
      String? videoUrl}) {
    return ShareItem._(
        content: url,
        label: label,
        subject: subject,
        shareType: ShareType.url,
        metadata: ShareMetadata(
          icon: iconUrl,
          videoUrl: videoUrl,
          caption: caption,
          imageUrl: imageUrl,
        ));
  }

  factory ShareItem.text(
    String text, {
    String? label,
    String? subject,
    String? imageUrl,
    String? videoUrl,
    String? icon,
    String? caption,
  }) {
    return ShareItem._(
        content: text,
        subject: subject,
        label: label,
        shareType: ShareType.text,
        metadata: ShareMetadata(
          imageUrl: imageUrl,
          videoUrl: videoUrl,
          icon: icon,
          caption: caption,
        ));
  }

  factory ShareItem.html(
    String html, {
    String? label,
    String? subject,
  }) {
    return ShareItem._(
        content: html,
        label: label,
        subject: subject,
        shareType: ShareType.html,
        metadata: const ShareMetadata.empty());
  }

  Map<String, dynamic>? toMap() {
    // ignore: unnecessary_cast
    return {
      'content': this.content,
      'label': this.label,
      'subject': this.subject,
      'shareType': this.shareType.index,
      'metadata': this.metadata?.toMap(),
      'variants': {
        for (var entry in variants.entries) entry.key: entry.value.toMap(),
      }
    }.prune();
  }

  ShareItem operator +(TargetedShareContent targeted) {
    return variant(targeted.targetAction, targeted.delegate);
  }

  ShareItem variant(TargetAction target, ShareContent content) {
    variants[target.key] = TargetedShareContent(content, target);
    return this;
  }
}

class TargetedShareContent {
  final ShareContent delegate;
  final TargetAction targetAction;

  TargetedShareContent(this.delegate, this.targetAction);

  dynamic get content => delegate.content;
  ShareType get shareType => delegate.shareType;

  Map<String, dynamic>? toMap() {
    return {
      'content': delegate.content,
      'shareType': delegate.shareType.index,
      'subject': delegate.subject,
      'targetAction': this.targetAction.toMap(),
    }.prune();
  }
}

class ShareMetadata {
  final String? icon;
  final String? imageUrl;
  final String? caption;
  final String? videoUrl;

  const ShareMetadata({
    this.icon,
    this.imageUrl,
    this.caption,
    this.videoUrl,
  });

  const ShareMetadata.empty()
      : icon = null,
        imageUrl = null,
        caption = null,
        videoUrl = null;

  Map<String, dynamic>? toMap() {
    return <String, dynamic>{
      'icon': this.icon,
      'imageUrl': this.imageUrl,
      'caption': this.caption,
      'videoUrl': this.videoUrl,
    }.prune();
  }
}

class ShareRequest {
  final List<ShareItem> items;
  final List<TargetAction> allowedActions;
  final List<TargetAction> restrictedActions;

  const ShareRequest({
    required this.items,
    this.allowedActions = const [],
    this.restrictedActions = const [],
  });

  Map<String, dynamic>? toMap() {
    return {
      'items': this.items.map((_) => _.toMap()).toList(),
      'allowedActions': this.allowedActions.map((_) => _.toMap()).toList(),
      'restrictedActions':
          this.restrictedActions.map((_) => _.toMap()).toList(),
    }.prune();
  }
}

class TargetAction {
  final String key;
  final String? iosName;
  final String? androidName;

  const TargetAction(this.key, {this.iosName, this.androidName});

  static const postToFacebook = TargetAction("postToFacebook");
  static const shareWithDevice = TargetAction("shareWithDevice");
  static const sendEmail = TargetAction("sendEmail");
  static const sendSms = TargetAction("sendSms");
  static const postToTencentWeibo = TargetAction("postToTencentWeibo");
  static const postToVimeo = TargetAction("postToVimeo");
  static const postToFlickr = TargetAction("postToFlickr");
  static const postToTwitter = TargetAction("postToTwitter");
  static const postToWeibo = TargetAction("postToWeibo");

  Map<String, dynamic>? toMap() {
    return {
      'key': this.key,
      'iosName': this.iosName,
      'androidName': this.androidName,
    }.prune();
  }
}

class ShareResponse {
  final String shareId;
  final String? activityType;
  final bool isCompleted;
  final Object error;
  final Object? result;

  const ShareResponse(
      {required this.shareId,
      this.activityType,
      required this.isCompleted,
      required this.error,
      this.result});

  factory ShareResponse.fromMap(dyn) {
    final map = dyn as Map<dynamic, dynamic>;
    return ShareResponse(
      shareId: map['shareId'] as String,
      activityType: map['activityType'] as String,
      isCompleted: map['isCompleted'] as bool,
      error: map['error'] as Object,
      result: map['result'] as Object,
    );
  }

  ShareResponse.timeout(this.shareId)
      : activityType = null,
        error = "Timed out",
        result = null,
        isCompleted = false;

  @override
  String toString() {
    return 'ShareResponse{shareId: $shareId, activityType: $activityType, isCompleted: $isCompleted, error: $error, result: $result}';
  }
}

extension _MapPrune on Map<String, dynamic>? {
  Map<String, dynamic>? prune() {
    final self = this;
    if (self == null) return null;
    self.removeWhere((k, v) => v == null);
    return self.isEmpty ? null : self;
  }
}
